"""add timescale

Revision ID: bab14faf0cd7
Revises: 12a0ab3826f3
Create Date: 2020-05-17 16:07:17.377521

"""

# revision identifiers, used by Alembic.
revision = 'bab14faf0cd7'
down_revision = '12a0ab3826f3'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # Extend the database with TimescaleDB
    op.execute('CREATE EXTENSION IF NOT EXISTS timescaledb CASCADE')


def downgrade():
    pass



import logging
import os

from peek_plugin_base.PeekVortexUtil import peekStorageName

logger = logging.getLogger(__name__)


class StorageTestMixin:
    def __init__(self):
        self._dbConn = None

    def setUp(self) -> None:
        from peek_storage._private.storage import setupDbConn
        from peek_storage._private.storage.DeclarativeBase import metadata
        import peek_storage
        from peek_plugin_base.storage.DbConnection import DbConnection
        from peek_storage._private.service.PeekStorageConfig import PeekStorageConfig

        from peek_platform import PeekPlatformConfig
        PeekPlatformConfig.componentName = peekStorageName

        config = PeekStorageConfig()

        alembicDir = os.path.join(
            os.path.dirname(peek_storage._private.__file__),
            "alembic")
        self._dbConn = DbConnection(dbConnectString=config.dbConnectString,
                                    metadata=metadata,
                                    alembicDir=alembicDir,
                                    dbEngineArgs=config.dbEngineArgs,
                                    enableCreateAll=False,
                                    enableForeignKeys=False)

        self._dbConn.migrate()

    def tearDown(self) -> None:
        self._dbConn.closeAllSessions()



"""add run_generic_python

Revision ID: 12a0ab3826f3
Revises: 34490abd765c
Create Date: 2020-04-25 15:28:50.279318

"""

# revision identifiers, used by Alembic.
revision = '12a0ab3826f3'
down_revision = '34490abd765c'
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    sql = '''
CREATE OR REPLACE FUNCTION peek_storage.run_generic_python(
    args_json character varying,
    kwargs_json character varying,
    class_method_to_run_str_ character varying,
    python_path character varying)
    RETURNS character varying
    LANGUAGE 'plpython3u'

    COST 100
    VOLATILE 
    
AS $BODY$

import json
from base64 import b64decode
args = json.loads(args_json)
kwargs = json.loads(kwargs_json)
classMethodToRunStr = class_method_to_run_str_
pythonPath = json.loads(python_path)

# ---------------
# Setup to use the virtual environment
import sys

sys.path.extend(pythonPath)

# ---------------
# Dynamically load the tuple create method

from importlib.util import find_spec, module_from_spec

modName, className, methodName = classMethodToRunStr.rsplit('.',2)

if modName in sys.modules:
    package_ = sys.modules[modName]

else:
    modSpec = find_spec(modName)
    if not modSpec:
         raise Exception("Failed to find package %s,"
                           " is the python package installed?" % modName)
    package_ = modSpec.loader.load_module()

Class_ = getattr(package_, className)
classMethodToRun = getattr(Class_, methodName)

result = classMethodToRun(plpy, *args, **kwargs)

return json.dumps({
    'result': result
})

$BODY$;

ALTER FUNCTION peek_storage.run_generic_python(character varying, character varying, character varying, character varying)
    OWNER TO peek;


          '''
    op.execute(sql)


def downgrade():
    sql = '''DROP FUNCTION peek_storage.run_generic_python(character varying, 
                                                            character varying, 
                                                            character varying, 
                                                            character varying);'''
    op.execute(sql)



"""add run worker plpy

Revision ID: 72518909970f
Revises: bab14faf0cd7
Create Date: 2020-05-20 20:13:19.092876

"""

# revision identifiers, used by Alembic.

revision = '72518909970f'
down_revision = 'bab14faf0cd7'
branch_labels = None
depends_on = None

from alembic import op

def upgrade():
    sql = '''
CREATE OR REPLACE FUNCTION peek_storage.run_worker_task_python(
    sqla_url character varying,
    args_json character varying,
    class_method_to_run_str_ character varying,
    python_path character varying)
    RETURNS character varying
    LANGUAGE 'plpython3u'

    COST 100
    VOLATILE

AS $BODY$

import json
from base64 import b64decode
sqlaUrl = sqla_url
argsJson = args_json
classMethodToRunStr = class_method_to_run_str_
pythonPath = json.loads(python_path)

# ---------------
# Setup Config

import sys

if not SD.get('peek_sys_path_set'):
    sys.path.extend(pythonPath)
    SD['peek_sys_path_set'] = True

# ---------------
# Setup to use the virtual environment

if SD.get('peek_config'):
    config = SD['peek_config']

else:
    from peek_plugin_base.PeekVortexUtil import peekWorkerName
    from peek_worker.PeekWorkerConfig import PeekWorkerConfig

    from peek_platform import PeekPlatformConfig
    PeekPlatformConfig.componentName = peekWorkerName

    config = PeekWorkerConfig()
    SD['peek_config'] = config

# ---------------
# Setup the workers logger

if not SD.get('peek_logger_setup'):
    import logging
    import os
    from peek_plugin_base.PeekVortexUtil import peekWorkerName
    from peek_platform.util.LogUtil import setupPeekLogger

    logging.root.setLevel(config.loggingLevel)

    setupPeekLogger("%s-%s" % (peekWorkerName, os.getpid()))
    SD['peek_logger_setup'] = True

# ---------------
# Initialise the DB Connections

if SD.get('CeleryDbConn'):
    CeleryDbConn = SD['CeleryDbConn']

else:
    from peek_plugin_base.worker import CeleryDbConn
    CeleryDbConn._dbConnectString = sqlaUrl
    CeleryDbConn._dbEngineArgs = {
        "max_overflow": 1,
        "pool_recycle": 600,
        "pool_size": 1,
        "pool_timeout": 0
    }

# ---------------
# Dynamically load the tuple create method

if SD.get(classMethodToRunStr):
    moduleMethodToRun = SD[classMethodToRunStr]

else:
    from importlib.util import find_spec, module_from_spec

    modName, methodName = classMethodToRunStr.rsplit('.',1)

    if modName in sys.modules:
        package_ = sys.modules[modName]

    else:
        modSpec = find_spec(modName)
        if not modSpec:
             raise Exception("Failed to find package %s,"
                               " is the python package installed?" % modName)
        package_ = modSpec.loader.load_module()

    moduleMethodToRun = getattr(package_, methodName)

# ---------------
# Load the arguments


if SD.get('peek_ArgTuple'):
    _RunPyInWorkerArgsTuple = SD['peek_ArgTuple']
    _RunPyInWorkerResultTuple = SD['peek_ResultTuple']
    json = SD['peek_json']

else:
    from peek_storage.plpython.RunWorkerTaskPyInPg \
        import _RunPyInWorkerArgsTuple, _RunPyInWorkerResultTuple

    SD['peek_ArgTuple'] = _RunPyInWorkerArgsTuple
    SD['peek_ResultTuple'] = _RunPyInWorkerResultTuple

    import json
    SD['peek_json'] = json

argsTuple = _RunPyInWorkerArgsTuple().fromJsonDict(json.loads(argsJson))

# ---------------
# Run the worker task

result = moduleMethodToRun(*argsTuple.args, **argsTuple.kwargs)

return json.dumps(_RunPyInWorkerResultTuple(result=result).toJsonDict())

$BODY$;

ALTER FUNCTION peek_storage.run_worker_task_python(character varying, character varying,
                                                   character varying, character varying)
    OWNER TO peek;


          '''
    op.execute(sql)


def downgrade():
    sql = '''DROP FUNCTION peek_storage.run_worker_task_python(character varying,
                                                               character varying,
                                                               character varying,
                                                               character varying);'''
    op.execute(sql)



from txhttputil.site.BasicResource import BasicResource

from peek_storage._private.service.sw_download.PeekSwDownloadResource import \
    PeekSwUpdateDownloadResource
from peek_storage._private.service.sw_download.PluginSwDownloadResource import \
    PluginSwDownloadResource

platformSiteRoot = BasicResource()


def setupPlatformSite():
    # Add the platform download resource
    platformSiteRoot.putChild(b'platform.sw_install.platform.download',
                              PeekSwUpdateDownloadResource())

    # Add the plugin download resource
    platformSiteRoot.putChild(b'platform.sw_install.plugin.download'
                              , PluginSwDownloadResource())



import logging
import unittest
from typing import List

from peek_storage._private.test.StorageTestMixin import StorageTestMixin
from peek_storage.plpython.RunPyInPg import runPyInPgBlocking

logger = logging.getLogger(__name__)


class RunPyInPgTestCase(unittest.TestCase, StorageTestMixin):
    def __init__(self, methodName='runTest'):
        unittest.TestCase.__init__(self, methodName)

        self._dbConn = None

    def setUp(self) -> None:
        StorageTestMixin.setUp(self)

    def tearDown(self) -> None:
        StorageTestMixin.tearDown(self)

    # -------------------------------------------------------------------------
    def test_args(self):
        result = runPyInPgBlocking(self._dbConn.ormSessionCreator,
                                   self._pg_args,
                                   None,
                                   ['SUC', 'CESS'])

        self.assertEqual(result, "SUCCESS")

    @classmethod
    def _pg_args(cls, plpy, successes: List[str]):
        return ''.join(successes)

    # -------------------------------------------------------------------------
    def test_kwwargs(self):
        result = runPyInPgBlocking(self._dbConn.ormSessionCreator,
                                   self._pg_kwargs,
                                   None,
                                   suc='SUC',
                                   cess='CESS KW')

        self.assertEqual(result, "SUCCESS KW")

    @classmethod
    def _pg_kwargs(cls, plpy, suc, cess):
        return suc + cess

    # -------------------------------------------------------------------------
    def test_pypg(self):
        result = runPyInPgBlocking(self._dbConn.ormSessionCreator,
                                   self._pg_plpy)

        self.assertEqual(result, "plpy exists")

    @classmethod
    def _pg_plpy(cls, plpy):
        assert plpy, "pypg is not defined"

        return "plpy exists"
    # -------------------------------------------------------------------------



import sys
from typing import Callable, Any, Optional

import json
from peek_plugin_base.storage.DbConnection import DbSessionCreator
from sqlalchemy import func
from vortex.Tuple import addTupleType, TupleField, Tuple

__sysPathsJson = json.dumps(sys.path)


@addTupleType
class _RunPyInPgArgTuple(Tuple):
    __tupleType__ = 'peek_storage._RunPyInPgArgTuple'
    args = TupleField()
    kwargs = TupleField()


@addTupleType
class _RunPyInPgResultTuple(Tuple):
    __tupleType__ = 'peek_storage._RunPyInPgResultTuple'
    result = TupleField()


def runPyInPgBlocking(dbSessionCreator: DbSessionCreator,
                      classMethodToRun: Callable,
                      classMethodToImportTuples: Optional[Callable],
                      *args,
                      **kwargs) -> Any:
    # noinspection PyProtectedMember
    argTupleJson = _RunPyInPgArgTuple(args=args, kwargs=kwargs)._toJson()

    loaderModuleClassMethodToRunStr = '.'.join([
        classMethodToRun.__self__.__module__,
        classMethodToRun.__self__.__name__,
        classMethodToRun.__name__
    ])

    loaderModuleClassMethodToImportStr = '.'.join([
        classMethodToImportTuples.__self__.__module__,
        classMethodToImportTuples.__self__.__name__,
        classMethodToImportTuples.__name__
    ]) if classMethodToImportTuples else 'None'

    session = dbSessionCreator()
    try:
        sqlFunc = func.peek_storage.run_generic_python(
            argTupleJson,
            loaderModuleClassMethodToRunStr,
            loaderModuleClassMethodToImportStr,
            __sysPathsJson
        )

        resultJsonStr: str = next(session.execute(sqlFunc))[0]
        session.commit()

        # noinspection PyProtectedMember
        resultTuple = _RunPyInPgResultTuple()._fromJson(resultJsonStr)

        return resultTuple.result

    finally:
        session.close()



import logging
from time import sleep

from peek_plugin_base.storage.DbConnection import DbConnection
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT, ISOLATION_LEVEL_DEFAULT

logger = logging.getLogger(__name__)


class StorageInit:
    def __init__(self, dbConnection: DbConnection):
        self._dbConnection = dbConnection

    def runPreMigrate(self):
        self._upgradeTimescaleDbExtension()

    def runPostMigrate(self):
        pass

    def _upgradeTimescaleDbExtension(self):
        rawConn = self._dbConnection.dbEngine.raw_connection()
        rawConn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        cursor = rawConn.cursor()
        try:

            logger.debug("Updating timescaledb extension")
            cursor.execute("ALTER EXTENSION timescaledb UPDATE")

            while len(rawConn.notices) < 2:
                sleep(2.0)


            else:
                logger.debug(rawConn.notices[1].replace('\n', ', '))

        except Exception as e:
            if 'Start a new session' in str(e):
                logger.debug("Skipping timescale extension update as the extension is"
                             " already loaded")

                return

            if 'extension "timescaledb" does not exist' in str(e):
                logger.debug("Skipping timescale extension update as the extension"
                             " doesn't exist yet")

                return


            logger.error("Updating timescaledb extesion failed")
            logger.exception(e)

        finally:
            rawConn.set_isolation_level(ISOLATION_LEVEL_DEFAULT)
            rawConn.close()
            cursor.close()



import json
import sys
from typing import Callable, Any

import json
from sqlalchemy import func
from vortex.Tuple import Tuple, addTupleType

from peek_plugin_base.storage.DbConnection import DbSessionCreator

__sysPathsJson = json.dumps(sys.path)


@addTupleType
class _RunPyInWorkerArgsTuple(Tuple):
    __tupleType__ = 'peek_storage.' + '_RunPyInWorkerArgsTuple'
    __slots__ = ('args', 'kwargs')


@addTupleType
class _RunPyInWorkerResultTuple(Tuple):
    __tupleType__ = 'peek_storage.' + '_RunPyInWorkerResultTuple'
    __slots__ = ('result',)


def runPyWorkerTaskInPgBlocking(dbSessionCreator: DbSessionCreator,
                                sqlaUrl: str,
                                moduleMethodToRun: Callable,
                                *argsTuple,
                                **kwargs) -> Any:
    argsTuple = _RunPyInWorkerArgsTuple(args=argsTuple, kwargs=kwargs)

    funcPath = moduleMethodToRun.__repr__().strip('<>').split(' ')[1]

    session = dbSessionCreator()
    try:
        sqlFunc = func.peek_storage.run_worker_task_python(
            sqlaUrl,
            json.dumps(argsTuple.toJsonDict()),
            funcPath,
            __sysPathsJson
        )

        resultJsonStr: str = next(session.execute(sqlFunc))[0]
        session.commit()

        resultTuple = _RunPyInWorkerResultTuple() \
            .fromJsonDict(json.loads(resultJsonStr))

        return resultTuple.result

    finally:
        session.close()



import sys
from collections import namedtuple
from typing import Callable, Dict, Optional

import json
from sqlalchemy import func
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import Select

from peek_plugin_base.storage.DbConnection import DbSessionCreator

LoadPayloadTupleResult = namedtuple("LoadPayloadTupleResult", ['count', 'encodedPayload'])

__sysPathsJson = json.dumps(sys.path)


def callPGLoadPayloadTuplesBlocking(dbSessionCreator: DbSessionCreator,
                                    sql: Select,
                                    sqlCoreLoadTupleClassmethod: Callable,
                                    payloadFilt: Optional[Dict] = None,
                                    fetchSize=50) -> LoadPayloadTupleResult:
    payloadFileJson = json.dumps(payloadFilt if payloadFilt else {})

    sqlStr = str(sql.compile(dialect=postgresql.dialect(),
                             compile_kwargs={"literal_binds": True}))

    loaderModuleClassMethodStr = '.'.join([
        sqlCoreLoadTupleClassmethod.__self__.__module__,
        sqlCoreLoadTupleClassmethod.__self__.__name__,
        sqlCoreLoadTupleClassmethod.__name__
    ])

    session = dbSessionCreator()
    try:
        sqlFunc = func.peek_storage.load_paylaod_tuples(
            sqlStr,
            payloadFileJson,
            loaderModuleClassMethodStr,
            __sysPathsJson,
            fetchSize
        )

        resultJsonStr: str = next(session.execute(sqlFunc))[0]

        resultJson: Dict = json.loads(resultJsonStr)
        if resultJson["encodedPayload"]:
            resultJson["encodedPayload"] = resultJson["encodedPayload"].encode()

        return LoadPayloadTupleResult(**resultJson)

    finally:
        session.close()
